#include <iostream>
#include <glut.h>
#include <stdint.h>
#include <list>
#include "glm/glm.hpp"
#include "glm/ext.hpp"

#include "render/render.hpp"
#include "scene_graph/scene_graph.hpp"
using namespace t3d;

const uint32_t SemanticMvp = 0;
const uint32_t SemanticTex = 1;

const uint32_t ModelNodeType = 1;

struct Vertex{
    glm::vec3 position;
    glm::vec3 normal;
    glm::vec2 texcoord;
    glm::vec4 joint;
    glm::vec4 weight;
};

struct Color{
    Color(uint8_t r = 255, uint8_t g = 255, uint8_t b = 255, uint8_t a = 255)
        : r(r), g(g), b(b), a(a) {}
    uint8_t r, g, b, a;
};

struct Material{
    Texture texture[8];
};

class ExampleNode : public Node{
public:
    void setModelMatrix(const glm::mat4 &modelMatrix){
        m_modelMatrix = modelMatrix;
    }

    const glm::mat4 &modelMatrix() const{
        return m_modelMatrix;
    }

protected:
    glm::mat4 m_modelMatrix;
};

class ModelNode : public ExampleNode{
public:
    virtual uint32_t type() const{
        return ModelNodeType;
    }

    void setVbo(const std::shared_ptr<Vbo> &vbo){
        m_vbo = vbo;
    }

    void setTexture(const std::shared_ptr<Texture> &texture){
        m_texture = texture;
    }

    void setShader(const std::shared_ptr<Shader> &shader){
        m_shader = shader;
    }

    const std::shared_ptr<Vbo> vbo() const{
        return m_vbo;
    }

    const std::shared_ptr<Shader> shader() const{
        return m_shader;
    }

    const std::shared_ptr<Texture> texture() const{
        return m_texture;
    }

private:
    std::shared_ptr<Vbo> m_vbo;
    std::shared_ptr<Texture> m_texture;
    std::shared_ptr<Shader> m_shader;
};

class ExampleRenderState : public RenderState{
public:
    ExampleRenderState()
        : projection(1.0), view(1.0), model(1.0)
    {}

    virtual RenderState *copy() const{
        ExampleRenderState *state = new ExampleRenderState();
        state->projection = projection;
        state->view = view;
        state->model = model;
        return state;
    }

    glm::mat4 projection, view, model;
};

class ModelNodeRenderDelegate : public NodeRenderDelegate{
public:
    virtual void render(const std::shared_ptr<Node> &node, const std::shared_ptr<RenderState> &renderState){
        std::shared_ptr<ModelNode> modelNode = std::static_pointer_cast<ModelNode>(node);
        std::shared_ptr<ExampleRenderState> exampleRenderState = std::static_pointer_cast<ExampleRenderState>(renderState);

        modelNode->vbo()->bind();
        modelNode->shader()->beginDraw();

        glm::mat4 mvp = exampleRenderState->projection * exampleRenderState->view * (exampleRenderState->model * modelNode->modelMatrix());
        modelNode->shader()->setUniformMatrix4fvArrayBySemantic(SemanticMvp, 1, &mvp[0][0]);
        modelNode->texture()->bind(0);
        modelNode->shader()->setUniform1iBySemantic(SemanticMvp, 0);
        glDrawElements(GL_TRIANGLES, modelNode->vbo()->indicesSize() / 2, GL_UNSIGNED_SHORT, 0);

        modelNode->shader()->endDraw();
    }
};

class ExampleRenderDelegate : public RenderDelegate{
public:
    virtual std::shared_ptr<RenderState> createChildRenderState(const std::shared_ptr<Node> &node, const std::shared_ptr<RenderState> &renderState){
        std::shared_ptr<ExampleNode> exampleNode = std::static_pointer_cast<ExampleNode>(node);
        std::shared_ptr<ExampleRenderState> exampleRenderState = std::static_pointer_cast<ExampleRenderState>(renderState);
        std::shared_ptr<ExampleRenderState> resultRenderState(static_cast<ExampleRenderState*>(renderState->copy()));
        resultRenderState->model *= exampleNode->modelMatrix();
        return resultRenderState;
    }
};

std::shared_ptr<Scene> scene;

void init(){
    std::string vertexSrc =
            "attribute vec4 position;\n"
            "attribute vec2 texcoord;\n"
            "uniform mat4 mvp;\n"
            "varying vec2 v_texcoord;\n"
            "void main(){\n"
            "   gl_Position = mvp * position;\n"
            "   v_texcoord = texcoord;\n"
            "}\n";

    std::string fragmentSrc =
            "uniform sampler2D texture;\n"
            "varying vec2 v_texcoord;\n"
            "void main(){\n"
            "   gl_FragColor = texture2D(texture, v_texcoord);\n"
            "}\n";

    glEnable(GL_DEPTH_TEST);

    std::shared_ptr<Texture> texture = std::make_shared<Texture>();
    Color bitmap[32 * 32];
    for (uint y = 0; y < 32; ++y)
        for (uint x = 0; x < 32; ++x){
            bitmap[y * 32 + x] = Color(float(x) / 32.0f * 255, float(y) / 32.0f * 255, 1.0 * 255, 1.0 * 255);
        }
    texture->loadFromData(bitmap, 32, 32, 32);

    std::shared_ptr<Shader> shader = std::make_shared<Shader>();
    shader->loadFromSources(vertexSrc, fragmentSrc);
    shader->appendAttribute(Attribute(shader->getAttributeHandle("position"), 3, 0, sizeof(Vertex)));
    shader->appendAttribute(Attribute(shader->getAttributeHandle("texcoord"), 2, 24, sizeof(Vertex)));
    shader->assignUniformSemanticByName("mvp", SemanticMvp);
    shader->assignUniformSemanticByName("texture", SemanticTex);

    std::shared_ptr<Vbo> vbo = std::make_shared<Vbo>();
    Vertex quadVertices[4];
    quadVertices[0].position = glm::vec3(-1, -1, 0); quadVertices[0].normal = glm::vec3(0, 0, -1); quadVertices[0].texcoord = glm::vec2(0, 0);
    quadVertices[1].position = glm::vec3(-1, 1, 0); quadVertices[1].normal = glm::vec3(0, 0, -1); quadVertices[1].texcoord = glm::vec2(0, 1);
    quadVertices[2].position = glm::vec3(1, 1, 0); quadVertices[2].normal = glm::vec3(0, 0, -1); quadVertices[2].texcoord = glm::vec2(1, 1);
    quadVertices[3].position = glm::vec3(1, -1, 0); quadVertices[3].normal = glm::vec3(0, 0, -1); quadVertices[3].texcoord = glm::vec2(1, 0);
    uint16_t quadIndices[6] = {0,1,2,0,2,3};
    vbo->loadFromData(&quadVertices[0], 4 * sizeof(Vertex), &quadIndices[0], 6 * sizeof(uint16_t));

    scene = std::make_shared<Scene>(std::make_shared<ExampleNode>());
    std::shared_ptr<ModelNode> model = std::make_shared<ModelNode>();
    model->setShader(shader);
    model->setTexture(texture);
    model->setVbo(vbo);
    scene->rootSceneNode()->addNode(model);

    std::shared_ptr<ExampleRenderDelegate> renderDelegate = std::make_shared<ExampleRenderDelegate>();
    std::shared_ptr<ModelNodeRenderDelegate> modelNodeRenderDelegate = std::make_shared<ModelNodeRenderDelegate>();
    scene->assignRenderDelegate(ModelNodeType, modelNodeRenderDelegate);
    scene->setRenderDelegate(renderDelegate);
}

void render(){
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

    std::shared_ptr<ExampleRenderState> renderState = std::make_shared<ExampleRenderState>();
    renderState->projection = glm::perspective(45.0f, 800.0f / 480.0f, 1.0f, 100.0f);
    renderState->view = glm::lookAt(glm::vec3(0, 0, -5), glm::vec3(0.0), glm::vec3(0, 1, 0));
    scene->render(renderState);

    glutSwapBuffers();
}

void reshape(int w, int h){

}

int main(int argc, char **argv)
{
    glutInit(&argc, argv);
    glutInitWindowSize(800, 480);
    glutInitDisplayMode(GLUT_RGB | GLUT_DOUBLE | GLUT_DEPTH);
    glutCreateWindow("Test");

    init();

    glutDisplayFunc(render);
    glutReshapeFunc(reshape);
    glutIdleFunc(render);

    glutMainLoop();
    return 0;
}

